#!/usr/bin/env python3

"""
Add N50 and size of bins to checkm2 results. 

:Example:
add_metrics_to_checkm.py -h
"""

# Metadata
__author__ = 'Mainguy Jean - Plateforme bioinformatique Toulouse'
__copyright__ = 'Copyright (C) 2022 INRAE'
__license__ = 'GNU General Public License'
__version__ = '0.1'
__email__ = 'support.bioinfo.genotoul@inra.fr'
__status__ = 'dev'


from argparse import ArgumentParser, ArgumentDefaultsHelpFormatter
import logging
from Bio import SeqIO
import os
import pandas as pd


def parse_arguments():
    """Parse script arguments."""
    parser = ArgumentParser(description="...",
                            formatter_class=ArgumentDefaultsHelpFormatter)

    parser.add_argument("-r", '--checkm2_result', required=True, help="result file made by checkm2.")
    parser.add_argument("-l", '--contig_to_length_tsv', required=True, help="Contig to length file.")
    parser.add_argument("-b", '--bin_dir', required=True, help="Bin set directory.")
    
    parser.add_argument("-o", "--outfile", default='checkm2_result_with_metrics.tsv' )
    parser.add_argument('-x', '--extension', help='Extension of input files. [Default: .fa]', default='.fa')

    parser.add_argument("-v", "--verbose", help="increase output verbosity",
                        action="store_true")

    args = parser.parse_args()
    return args


def compute_N50(list_of_lengths):
    """Calculate N50 for a sequence of numbers.

    Args:
        list_of_lengths (list): List of numbers.

    Returns:
        int: N50 value.

    """

    list_of_lengths = sorted(list_of_lengths)

    sum_len = sum(list_of_lengths)

    cum_length = 0
    for length in list_of_lengths:
        if cum_length + length >= sum_len/2:
            return length
        cum_length += length
    return length


def get_bin_composition(fasta_bins):

    bin2contigs = {}

    for seq_file in fasta_bins:
        bin2contigs['.'.join(os.path.basename(seq_file).split('.')[:-1])] = [record.id for record in SeqIO.parse(seq_file, "fasta")]

    return bin2contigs


def get_bin_N50(bin_id, bin2contigs, df_len):
    contigs = bin2contigs[(str(bin_id))]
    lengths = df_len.loc[contigs, "length"]
    return compute_N50(lengths)


def get_bin_size(bin_id, bin2contigs, df_len):
    contigs = bin2contigs[(str(bin_id))]
    return df_len.loc[contigs, "length"].sum()


def main():

    args = parse_arguments()

    if args.verbose:
        logging.basicConfig(format="%(levelname)s: %(message)s", level=logging.DEBUG)
        logging.info('Mode verbose ON')

    else:
        logging.basicConfig(format="%(levelname)s: %(message)s")
    
    
    contig_to_length_tsv = args.contig_to_length_tsv
    bin_dir = args.bin_dir
    fasta_extension = args.extension

    checkm2_result = args.checkm2_result

    logging.info(f"Parsing {contig_to_length_tsv}.")

    df_len = pd.read_csv(contig_to_length_tsv, sep="\t", index_col="seqid")

    logging.info("Get bin composition by parsing bin fasta files.")

    bin_files = [os.path.join(bin_dir, seq_file) for seq_file in os.listdir(bin_dir) if seq_file.endswith(fasta_extension)]

    bin2contigs = get_bin_composition(bin_files)

    logging.info("Parse checkM2 results")
    df_checkm = pd.read_csv(checkm2_result, sep="\t")

    df_checkm["N50"] = df_checkm["Name"].apply(
        lambda x: get_bin_N50(x, bin2contigs, df_len)
    )
    df_checkm["Size"] = df_checkm["Name"].apply(
        lambda x: get_bin_size(x, bin2contigs, df_len)
    )

    # metawrap identifies header with completeness in lower case
    df_checkm = df_checkm.rename(columns={"Name":"genome", "Completeness": "completeness", "Contamination": "contamination"})

    df_checkm.to_csv(args.outfile, sep="\t", index=False)


if __name__ == "__main__":
    main()
