#!/usr/bin/env python3

"""
This script takes multiple folders containing bins from different methods and their metrics files generated by CheckM. 
Then start by matching corresponding bins in the two first sets based on a minimum of 80% overlap (by length) of the bins and decides which of the two
bin versions is best in that particular bin, making a new bins set. 
This new bin set is then used in the same way with the third bins sets.

Then it makes a new folder into which it puts the best version of each bin (changing the naming in
the process), and also makes a new .stats file which is consistant with the new bin folder.

:Example:
consolidate_multiple_bins_sets.py --bin_dirs binsA binsAB binsABC --bin_stats binsABC.stats binsAB.stats binsA.stats -l contig_to_length.tsv
"""

# Metadata
__author__ = "Mainguy Jean - Plateforme bioinformatique Toulouse"
__copyright__ = "Copyright (C) 2022 INRAE"
__license__ = "GNU General Public License"
__version__ = "0.1"
__email__ = "support.bioinfo.genotoul@inra.fr"
__status__ = "dev"


from argparse import ArgumentParser, ArgumentDefaultsHelpFormatter
import logging
import os
import pandas as pd
from Bio import SeqIO
from collections import defaultdict
import shutil
from functools import reduce
import sys
pd.set_option('display.max_rows', 200)

def parse_arguments():
    """Parse script arguments."""
    parser = ArgumentParser(
        description="...", formatter_class=ArgumentDefaultsHelpFormatter
    )

    parser.add_argument("--bin_stats", required=True, nargs="+", 
                            help="Stat file on bins sets.")
    parser.add_argument("--bin_dirs", required=True,
                        nargs="+", help="Directories of bins sets.")

    parser.add_argument("-l", '--contig_to_length_tsv',
                        required=True, help="Contig to length file.")

    parser.add_argument('-o', "--out_bin", default="consolidated_bins", help="Consolidated final bins set name. Create a directory <out_bin> and a tsv with bins stat <out_bin>.stats")

    parser.add_argument("--min_completeness", default=50, type=int, help="Bin minimum completeness treshold.")

    parser.add_argument("--max_contamination", default=10, type=int, help="Bin maximum contamination threshold.")
    parser.add_argument("--min_overlap", default=80, type=int, help="Minimum overlap percentage (by length) between two bins to consider them corresponding.")

    parser.add_argument(
        "-v", "--verbose", help="increase output verbosity", action="store_true"
    )

    args = parser.parse_args()
    return args


def get_bin_contigs(seqfile):
    return {record.id for record in SeqIO.parse(seqfile, "fasta")}


def get_bin_file(bin_dir, bin_id):
    bin_file = os.path.join(bin_dir, f"{bin_id}.fa")
    if not os.path.isfile(bin_file):
        raise FileNotFoundError(f"Bin file {bin_id}.fa does not exists in {bin_dir}")
    return bin_file


def parse_bin_set_stats(bin_dir, stat_file, min_complet, max_conta):
    logging.info(f"Parsing {stat_file} and {bin_dir}")

    df = pd.read_csv(stat_file, sep="\t")

    df["genome"] = df["genome"].astype(str)

    df["bin_file"] = df["genome"].apply(lambda x: get_bin_file(bin_dir, x))

    df = df.set_index("bin_file", drop=False)

    # # keep only good bin.
    good_bin_filt = (df["completeness"] >= min_complet) & (
        df["contamination"] <= max_conta
    )
    df = df.loc[good_bin_filt]

    # score each bin

    df["score"] = df["completeness"] - df["contamination"] * 5

    # get contigs of each bin

    df["contigs"] = df["bin_file"].apply(lambda x: get_bin_contigs(x))

    df['origin'] = bin_dir

    return df


def check_input_files_consistency(bin_dirs, stat_files):
    logging.info("Checking input files consistency.")
    if len(bin_dirs) != len(stat_files):
        logging.error(
            f"Number of bin dirs ({len(bin_dirs)}) and stat files ({len(stat_files)}) given to the script are different."
        )
        exit(1)

    bin_name2bin_dir = {
        os.path.basename(os.path.normpath(bin_dir)): bin_dir for bin_dir in bin_dirs
    }
    bin_dir_stat_file_tuples = []

    for stat_file in stat_files:

        bins_set_name = ".".join(os.path.basename(stat_file).split(".")[:-1])
        try:
            bin_dir = bin_name2bin_dir[bins_set_name]
        except KeyError:
            raise FileNotFoundError(
                f"Stat file {stat_file} has no matching bin dir. Expecting bin dir basename: {bins_set_name}"
            )

        if not os.path.isdir(bin_dir):
            raise FileNotFoundError(f"Bin dir does not exists: {bin_dirs}.")

        bin_dir_stat_file_tuples.append((bin_dir, stat_file))

    return bin_dir_stat_file_tuples


def consolidate_two_bins_sets(df_bins_1, df_bins_2, df_len, min_overlap):
    df_bins_1 = df_bins_1[~df_bins_1.index.duplicated(keep='first')]
    df_bins_2 = df_bins_2[~df_bins_2.index.duplicated(keep='first')]
    #logging.info(f"consolidate {set(df_bins_1['origin'])} with {set(df_bins_2['origin'])} ")

    df_bins = pd.concat([df_bins_1, df_bins_2])

    bin_pairs, lonely_bins_2 = get_overlapping_bins(df_bins_1, df_bins_2, df_len, min_overlap )

    selected_bins = []
    # go through all overlapping bins and chose the best ones
    for bin_1, bins_2 in bin_pairs.items():

        index_bins = [bin_1] + bins_2
        if len(bins_2) > 1:
            logging.info(f"The bin {bin_1} overlap with multiple bins: {bins_2}")

        df_bin_group = df_bins.loc[index_bins]

        selected_bin = df_bin_group.loc[df_bin_group['score'].idxmax()]
        selected_bins.append(selected_bin)
        

    selected_bins += [df_bins.loc[bin_file] for bin_file in lonely_bins_2]
    df_select = pd.DataFrame(selected_bins)
    return df_select

def get_overlapping_bins(bins_1, bins_2, df_len, min_overlap ):

    bin_pairs = {}
    set2_paired_bins = set()

    for index1, row1 in bins_1.iterrows():
        bin_pairs[index1] = []
        for index2, row2 in bins_2.iterrows():

            shared_contigs = row1['contigs'] & row2['contigs']

            if len(shared_contigs) == 0:
                 continue

            bin1_size =  df_len.loc[list(row1['contigs']), 'length'].sum()
            bin2_size =  df_len.loc[list(row2['contigs']), 'length'].sum()

            shared_length = df_len.loc[list(shared_contigs), 'length'].sum()

            # chose the highest % ID, depending of which bin is a subset of the other
            bin_overlaping_prct = 100 * max([shared_length/bin1_size, shared_length/bin2_size])

            if bin_overlaping_prct >= min_overlap:

                bin_pairs[index1].append(index2)

                set2_paired_bins.add(index2)
                
    logging.info(f"There are {len(bins_1)} bins in bins1 and {len(bin_pairs)} of them are pairing with bins_2")
    
    logging.info(f"There are  {len(bins_2)} bins in bins2 and {len(set2_paired_bins)} of them  are pairing with bins1")

    lonely_bins_2 = set(bins_2.index) -  set2_paired_bins 
    return bin_pairs, lonely_bins_2

def main():

    args = parse_arguments()

    if args.verbose:
        logging.basicConfig(format="%(levelname)s: %(message)s", level=logging.DEBUG)
        logging.info("Mode verbose ON")

    else:
        logging.basicConfig(format="%(levelname)s: %(message)s")

    min_complet = args.min_completeness
    max_conta =  args.max_contamination

    min_overlap = 80

    contig_to_length_tsv = args.contig_to_length_tsv

    checkm_results = (args.bin_stats)
    bin_dirs = args.bin_dirs


    out_bin = args.out_bin
    out_bin_stats = f"{out_bin}.stats"
    os.makedirs(out_bin, exist_ok=True)

    df_len = pd.read_csv(contig_to_length_tsv, sep='\t', index_col="seqid")
    
    bin_dir_stat_file_tuples = check_input_files_consistency(bin_dirs, checkm_results)

    logging.info(f'Consolidating {len(bin_dir_stat_file_tuples)} bins sets.')

    df_bins_iter = (
        parse_bin_set_stats(bin_dir, stat, min_complet, max_conta)
        for bin_dir, stat in bin_dir_stat_file_tuples
    )

    final_bins_set_stat = reduce(lambda b1, b2 : consolidate_two_bins_sets(b1, b2, df_len, min_overlap), df_bins_iter)
    if len(final_bins_set_stat) == 0:
        logging.warning(f"None of the bins meet minimum completeness and maximun contamination thresholds")
        sys.exit(255)

    logging.info('Writting final bins set ')

    final_bins_set_stat['genome']  = 'bins.' + final_bins_set_stat.reset_index().index.astype(str)
    
    # cleaning bin stat columns
    final_bins_set_stat = final_bins_set_stat.drop(["bin_file", "score", "contigs"], axis=1)
    final_bins_set_stat = final_bins_set_stat[~final_bins_set_stat.index.duplicated(keep='first')]

    logging.info(f'Writing final bin set in : {out_bin}')
    for bin_file in final_bins_set_stat.index:

        new_bin_id =  final_bins_set_stat.loc[bin_file, "genome"]
        outfile = os.path.join(out_bin, f"{new_bin_id}.fa")
        shutil.copyfile(bin_file, outfile)

    logging.info(f'Writing final bin set stats: {out_bin_stats}')
    final_bins_set_stat.to_csv(out_bin_stats, sep="\t", index=False)

if __name__ == "__main__":
    main()
