#!/usr/bin/env python3

"""
From a list of directory of sequences fasta files, get a uniq file containing all sequences once.

:Example:
python template.py -v
"""

# Metadata
__author__ = 'Mainguy Jean - Plateforme bioinformatique Toulouse'
__copyright__ = 'Copyright (C) 2022 INRAE'
__license__ = 'GNU General Public License'
__version__ = '0.1'
__email__ = 'support.bioinfo.genotoul@inra.fr'
__status__ = 'dev'


from argparse import ArgumentParser, ArgumentDefaultsHelpFormatter
import logging
from Bio import SeqIO
import os


def parse_arguments():
    """Parse script arguments."""
    parser = ArgumentParser(description="...",
                            formatter_class=ArgumentDefaultsHelpFormatter)

    parser.add_argument('--seq_dirs', nargs='+')
    parser.add_argument("-o", "--outfile", default='uniq_concat_sequences.fasta' )
    parser.add_argument("-v", "--verbose", help="increase output verbosity",
                        action="store_true")
    parser.add_argument('-x', '--extension', help='Extension of input files. [Default: .fa]', default='.fa')

    args = parser.parse_args()
    return args


def main():

    args = parse_arguments()

    if args.verbose:
        logging.basicConfig(format="%(levelname)s: %(message)s", level=logging.DEBUG)
        logging.info('Mode verbose ON')

    else:
        logging.basicConfig(format="%(levelname)s: %(message)s")
    

    fasta_extension = args.extension

    all_seq_files = []
    for seq_dir in args.seq_dirs:
        all_seq_files += [os.path.join(seq_dir, seq_file) for seq_file in os.listdir(seq_dir) if seq_file.endswith(fasta_extension)]
    
    logging.info(f"Found {len(all_seq_files)}  seq files to parse.")
    
    seqs_processed = []
    with open(args.outfile, 'w') as f_out:
        for seq_file in all_seq_files:
            
            with open(seq_file, "r") as handle:
                for record in SeqIO.parse(handle, "fasta"):
                    if record.id not in seqs_processed:
                        
                        SeqIO.write(record, f_out, "fasta")
                        seqs_processed.append(record.id)

    logging.info(f'Have written {len(seqs_processed)} uniq sequences in {args.outfile}')

if __name__ == '__main__':
    main()
