#!/usr/bin/env python3

"""
Dereplicate contigs in bins. 

When a contig is found in multiple bins, it is kept only in the best bin or it is remove from the bins set when --remove flag is on.

:Example:
dereplicate_contigs_in_bins.py  --bins_stat binsM.stats  --bin_dir binsM --out_bin_dir binsO
"""

# Metadata
__author__ = "Mainguy Jean - Plateforme bioinformatique Toulouse"
__copyright__ = "Copyright (C) 2022 INRAE"
__license__ = "GNU General Public License"
__version__ = "0.1"
__email__ = "support.bioinfo.genotoul@inra.fr"
__status__ = "dev"


from argparse import ArgumentParser, ArgumentDefaultsHelpFormatter
import logging
from Bio import SeqIO
import os
import pandas as pd
import sys
from collections import defaultdict


def parse_arguments():
    """Parse script arguments."""
    parser = ArgumentParser(
        description="...", formatter_class=ArgumentDefaultsHelpFormatter
    )

    parser.add_argument(
        "-r",
        "--bins_stat",
        required=True,
        help="bin set stat file with completeness, contamination and N50 metrics.",
    )

    parser.add_argument("-b", "--bin_dir", required=True, help="Bin set directory.")

    parser.add_argument(
        "-o",
        "--out_bin_dir",
        required=True,
    )
    parser.add_argument(
        "-x",
        "--extension",
        help="Extension of input files. [Default: fa]",
        default="fa",
    )

    parser.add_argument(
        "--remove",
        help="Remove contigs that end up in more than one bin",
        action="store_true",
    )

    parser.add_argument(
        "-v", "--verbose", help="increase output verbosity", action="store_true"
    )

    args = parser.parse_args()
    return args


def get_contig2bins(fasta_bins):

    contig2bins = defaultdict(list)

    for seq_file in fasta_bins:
        bin_id = ".".join(os.path.basename(seq_file).split(".")[:-1])
        for record in SeqIO.parse(seq_file, "fasta"):
            contig2bins[record.id].append(bin_id)

    return contig2bins


def main():

    args = parse_arguments()

    if args.verbose:
        logging.basicConfig(format="%(levelname)s: %(message)s", level=logging.DEBUG)
        logging.info("Mode verbose ON")

    else:
        logging.basicConfig(format="%(levelname)s: %(message)s")

    bins_stats = args.bins_stat
    bin_dir = args.bin_dir
    out_bin_dir = args.out_bin_dir
    fasta_extension = args.extension
    remove_flag = args.remove

    os.makedirs(out_bin_dir, exist_ok=True)

    logging.info("Loading bin completion and contamination scores...")
    df = pd.read_csv(bins_stats, sep="\t", index_col="genome")
    df["score"] = df["completeness"] - 5 * df["contamination"] + 1e-10 * df["N50"]



    logging.info("Loading in contigs in each bin...")

    bin_files = [
        os.path.join(bin_dir, seq_file)
        for seq_file in os.listdir(bin_dir)
        if seq_file.endswith(fasta_extension)
    ]

    contig2bins = get_contig2bins(bin_files)


    logging.info("Get dereplicated contig list per bin...")

    bin2uniqcontigs = defaultdict(list)

    for contig, bins in contig2bins.items():
        if len(bins) > 1 and remove_flag:
            logging.info(
                f"{contig} is removed because it is found in more than one bin. (remove flag is on)"
            )
            continue

        best_bin = df.loc[bins, "score"].idxmax()

        bin2uniqcontigs[best_bin].append(contig)

    # go over the bin files again and make a new dereplicated version of each bin file
    logging.info(
        f"Making a new dereplicated version of each bin file and writting them in {out_bin_dir}"
    )
    for bin_id, contigs in bin2uniqcontigs.items():

        bin_file = os.path.join(bin_dir, f"{bin_id}.{fasta_extension}")
        out_bin_file = os.path.join(out_bin_dir, f"{bin_id}.{fasta_extension}")

        with open(bin_file) as in_fh, open(out_bin_file, "w") as out_fh:
            for record in SeqIO.parse(in_fh, "fasta"):
                if record.id in contigs:
                    SeqIO.write(record, out_fh, "fasta")


if __name__ == "__main__":
    main()
