#!/usr/bin/env python3

"""
Get lengths of contigs

:Example:
python template.py -v
"""

# Metadata
__author__ = 'Mainguy Jean - Plateforme bioinformatique Toulouse'
__copyright__ = 'Copyright (C) 2022 INRAE'
__license__ = 'GNU General Public License'
__version__ = '0.1'
__email__ = 'support.bioinfo.genotoul@inra.fr'
__status__ = 'dev'


from argparse import ArgumentParser, ArgumentDefaultsHelpFormatter
import logging
from Bio import SeqIO
import os


def parse_arguments():
    """Parse script arguments."""
    parser = ArgumentParser(description="...",
                            formatter_class=ArgumentDefaultsHelpFormatter)

    parser.add_argument('--seq_dirs', nargs='+')
    parser.add_argument("-o", "--outfile", default='contig_to_length.tsv' )
    parser.add_argument("-v", "--verbose", help="increase output verbosity",
                        action="store_true")
    parser.add_argument('-x', '--extension', help='Extension of input files. [Default: .fa]', default='.fa')

    args = parser.parse_args()
    return args


def main():

    args = parse_arguments()

    if args.verbose:
        logging.basicConfig(format="%(levelname)s: %(message)s", level=logging.DEBUG)
        logging.info('Mode verbose ON')

    else:
        logging.basicConfig(format="%(levelname)s: %(message)s")
    

    fasta_extension = args.extension

    seqid2length = {}
    all_seq_files = []
    for seq_dir in args.seq_dirs:
        all_seq_files += [os.path.join(seq_dir, seq_file) for seq_file in os.listdir(seq_dir) if seq_file.endswith(fasta_extension)]
    
    logging.info(f"Found {len(all_seq_files)}  seq files to parse.")
    
    for seq_file in all_seq_files:
        with open(seq_file, "r") as handle:
            for record in SeqIO.parse(handle, "fasta"):
                if record.id not in seqid2length:
                    seqid2length[record.id] = len(record.seq)
                    
    seqid2length_iter = (f"{seqid}\t{length}" for seqid, length in seqid2length.items())
    with open(args.outfile, 'w') as fl:
        fl.write(f"seqid\tlength\n")
        fl.write('\n'.join(seqid2length_iter))

if __name__ == '__main__':
    main()
