#!/usr/bin/env python3

"""
Setup intermediate files of checkm2 of bin sets 
to skip prodigal and diamond steps in checkm2

:Example:
python template.py -v
"""

# Metadata
__author__ = 'Mainguy Jean - Plateforme bioinformatique Toulouse'
__copyright__ = 'Copyright (C) 2022 INRAE'
__license__ = 'GNU General Public License'
__version__ = '0.1'
__email__ = 'support.bioinfo.genotoul@inra.fr'
__status__ = 'dev'


from argparse import ArgumentParser, ArgumentDefaultsHelpFormatter
import logging
from Bio import SeqIO
import os
from collections import defaultdict
import os
import errno
import sys
import pandas as pd

def parse_arguments():
    """Parse script arguments."""
    parser = ArgumentParser(description="...",
                            formatter_class=ArgumentDefaultsHelpFormatter)

    parser.add_argument('-b', '--bin_dir', required=True, help="Bin directory.",)
    parser.add_argument('-f', '--faa_raw', required=True, help="Protein sequences predicted with prodigal on all contigs.",)
    parser.add_argument('-d', '--diamond_raw', required=True, help="Results from diamond launched on all contigs.",)
    parser.add_argument("-o", "--outdir", required=True)
    parser.add_argument("-v", "--verbose", help="increase output verbosity",
                        action="store_true")
    parser.add_argument('-x', '--extension', help='Extension of input files. [Default: .fa]', default='.fa')

    args = parser.parse_args()
    return args

def get_contig2bin(fasta_bins):

    contig2bin = {}

    for seq_file in fasta_bins:
        bin_id = '.'.join(os.path.basename(seq_file).split('.')[:-1])
        contig2bin.update({record.id:bin_id for record in SeqIO.parse(seq_file, "fasta")}) 

    return contig2bin

def make_sure_path_exists(path):
    """
    Create directory if it does not exist.
    
    Taken from Checkm2. https://github.com/chklovski/CheckM2/blob/e563159f005344d798defb60cea4189e0be8ec92/checkm2/fileManager.py#L218
    """
    
    if not path:
        return

    try:
        os.makedirs(path)
    except OSError as exception:
        if exception.errno != errno.EEXIST:
            logging.error('Specified path does not exist: ' + path + '\n')
            sys.exit(1)

def from_faa_to_bin(faa_id, contig2bin):
    contig = "_".join(faa_id.split('_')[:-1])
    
    try:
        bin_id = contig2bin[contig]
    except KeyError:
        bin_id = None
        
    return bin_id

def get_bin2faa_seq(faa_file, contig2bin):

    bin2faa_seq  = defaultdict(list)
    contigs_not_in_bin = set()
    # load all faa sequences into memory and make a bin2sequencerecord
    for record in SeqIO.parse(faa_file, "fasta"):
        
        contig = '_'.join(record.id.split('_')[:-1])

        if contig not in contig2bin:
            contigs_not_in_bin.add(contig)
            continue

        bin_id = contig2bin[contig]
        bin2faa_seq[bin_id].append(record)
        
    logging.info(f"{len(contigs_not_in_bin)} contigs have been ignored as they do not belong to any bins.")
    return bin2faa_seq

def main():

    args = parse_arguments()

    if args.verbose:
        logging.basicConfig(format="%(levelname)s: %(message)s", level=logging.DEBUG)
        logging.info('Mode verbose ON')

    else:
        logging.basicConfig(format="%(levelname)s: %(message)s")

    bin_dir = args.bin_dir
    fasta_extension = args.extension

    diamond_outname = "DIAMOND_RESULTS.tsv"

    diamond_raw = args.diamond_raw

    prodigal_outdir_name = "protein_files"
    diamond_outdir_name = "diamond_output"
    
    diamond_header_separator = 'Ω'
    
    prodigal_out_path = os.path.join(args.outdir, prodigal_outdir_name)
    diamond_out_path = os.path.join(args.outdir, diamond_outdir_name)

    make_sure_path_exists(prodigal_out_path)
    make_sure_path_exists(diamond_out_path)


    diamond_output = os.path.join(diamond_out_path, diamond_outname)


    logging.info("Get bin composition by parsing bin fasta files.")

    bin_files = [os.path.join(bin_dir, seq_file) for seq_file in os.listdir(bin_dir) if seq_file.endswith(fasta_extension)]
    contig2bin = get_contig2bin(bin_files)
    
    logging.info("Setting up faa output.")

    bin2faa_seq = get_bin2faa_seq(args.faa_raw, contig2bin)

    logging.info(f"Writting faa into {len(bin2faa_seq)} bin files in {prodigal_out_path}.")

    for bin_id, faa_seq in bin2faa_seq.items():

        bin_out_file = os.path.join(prodigal_out_path, f'{bin_id}.faa')
        SeqIO.write(faa_seq, bin_out_file, "fasta")

    logging.info("Setting up DIAMOND output.")
    
    headers = ['query_id', "subject_id"] + [f"_{i}" for i in range(10)]
    df = pd.read_csv(diamond_raw, sep='\t', names= headers)

    # Get bin id for each query id
    df['bin'] = df['query_id'].apply(lambda x: from_faa_to_bin(x,contig2bin))

    # remove line that are from query that does not belong to current bin set
    df = df.loc[ ~ df['bin'].isna()]

    # concat binid and query id with checkm2 separator 
    df['query_id'] = df['bin'] + diamond_header_separator + df['query_id']
    df = df.drop('bin', axis=1)

    df.to_csv(diamond_output, sep="\t", index=False, header=False)

    return

if __name__ == '__main__':
    main()
