#!/usr/bin/env python

"""--------------------------------------------------------------------
  Script Name: from_command_to_process.py
  Description: Take and parse a raw nextflow command as well as a process
  yaml file giving information on each process to output expected process 
  launched. 
  Input files: metagwgs nextflow command & process.yaml
  Created By:  Jean Mainguy
  Date:        2022-11-30
-----------------------------------------------------------------------
"""

# Metadata.
__author__ = "Jean Mainguy"
__copyright__ = "Copyright (C) 2023 INRAE"
__license__ = "GNU General Public License"
__email__ = "support.bioinfo.genotoul@inra.fr"
__status__ = "dev"


# Modules importation
from argparse import ArgumentParser, ArgumentDefaultsHelpFormatter
import logging
import yaml


def parse_arguments():
    """Parse script arguments."""
    parser = ArgumentParser(
        description="Generates density plot distribution of kiaju match length",
        formatter_class=ArgumentDefaultsHelpFormatter,
    )

    parser.add_argument(
        "-c", "--command", required=True, help="a full nextflow command in quote. Arguments should be explicit."
    )

    parser.add_argument("-p", "--process_yaml", default="process.yaml file with info on all process.")

    parser.add_argument("-v", "--verbose", help="increase output verbosity", action="store_true")

    args = parser.parse_args()
    return args


def parse_nextflow_command(command):
    command_split = command.split()
    print(command_split)
    key_val_args = {}
    flags = set()
    pass_next_key_val = False
    for key, value in zip(command_split, command_split[1:] + ["<END>"]):
        if pass_next_key_val:
            pass_next_key_val = False
            continue

        if key.startswith("--") and not value.startswith("-") and value != "<END>":
            key = key[2:]
            key_val_args[key] = value
            pass_next_key_val = True

            logging.info(f"Pair: {key} = {value}")

        elif key.startswith("--") and (value.startswith("-") or value == "<END>"):
            flags.add(key[2:])
            logging.info(f"Flag: {key}")
        else:
            print(f"'{key}' argument is ignored. it is likely a nextflow specific parameter.")

    return flags, key_val_args


def parse_process_yaml(yaml_process):

    with open(yaml_process) as stream:
        process_info = yaml.safe_load(stream)

    logging.info(process_info.keys())

    step_order = process_info["step_order"]
    logging.info(f"Step order {step_order}")

    process_to_labels = process_info["process"]
    return step_order, process_to_labels


def main():

    args = parse_arguments()

    if args.verbose:
        logging.basicConfig(format="%(levelname)s: %(message)s", level=logging.DEBUG)
        logging.info("Mode verbose ON")

    else:
        logging.basicConfig(format="%(levelname)s: %(message)s")

    yaml_process = args.process_yaml
    command = args.command.lower()

    mandatory_arguments = ["type"]  # , "--assembly_tool"]

    # Parse nextflow command
    flags, key_val_args = parse_nextflow_command(command)

    # Check command
    for arg in mandatory_arguments:
        print(key_val_args)
        print(key_val_args.keys())
        if arg not in key_val_args.keys():
            logging.error(
                f"Error {arg} should be in command. \
                Please be explicit! All this arguments '{mandatory_arguments}' should be found in the command."
            )
            exit(1)
    # Parse process yaml file
    logging.info(f"Parse process yaml file: {yaml_process}")

    step_order, process_to_labels = parse_process_yaml(yaml_process)

    # Parse command
    # Arguments are cleaned
    # Arguments that starts with stop_at and skip_ are
    # spotted and put appart to be used later

    stop_at_flags = {flag for flag in flags if flag.startswith("stop_at_")}
    assert len(stop_at_flags) == 1, "More than one stop_at flag found in command. Please just use one."
    stop_at = stop_at_flags.pop()

    skip_params = {flag for flag in flags if flag.startswith("skip_")}

    other_flags = flags - (stop_at_flags | skip_params)

    logging.info(f"Skip params: {skip_params}")
    logging.info(f"Stop_at param: {stop_at}")
    logging.info(f"Other flags: {other_flags}")

    # determine which step is applied according the stop_at argument if any

    do_step_flag = True
    steps_to_do = set()
    for step in step_order:

        if do_step_flag:
            steps_to_do.add(step)
        else:
            skip_params.add(step)

        if step == stop_at:
            do_step_flag = False

    params = steps_to_do | other_flags | set(key_val_args.values()) | set(key_val_args.keys())

    # Get trough all process and check if they are expected are not according params
    unlaunched_process = set()
    launched_process = set()

    for process, labels in process_to_labels.items():
        logging.info(f"{process} --> {labels}")

        try:
            required_labels = {l.lower() for l in labels["needs"]}
        except KeyError:
            required_labels = []

        try:
            skip_labels = [l.lower() for l in labels["skip_by"]]
        except KeyError:
            skip_labels = []

        process_skip_params = [l for l in required_labels if l in skip_params]
        process_skip_params += [l for l in skip_labels if l in params]
        if process_skip_params:
            logging.info(f"{process} is skipped because it of these parameters: {skip_params}")
            continue

        missing_params = [label for label in required_labels if label not in params]

        if missing_params:
            logging.info(
                f"{process} is not launched because some of its required parameters are missing: {missing_params}"
            )
            unlaunched_process.add(process)
        else:
            logging.info(f"{process} is expected to be launched. All its needs are fullfiled")
            launched_process.add(process)

    logging.info(launched_process)


if __name__ == "__main__":
    main()
